aubiodefvalue = {
    # we have some clean up to do
    'buf_size': 'Py_default_vector_length',
    'win_s': 'Py_default_vector_length',
    'size': 'Py_default_vector_length',
    # and here too
    'hop_size': 'Py_default_vector_length / 2',
    'hop_s': 'Py_default_vector_length / 2',
    # these should be alright
    'samplerate': 'Py_aubio_default_samplerate',
    # now for the non obvious ones
    'n_filters': '40',
    'n_coeffs': '13',
    'nelems': '10',
    'flow': '0.',
    'fhig': '1.',
    'ilow': '0.',
    'ihig': '1.',
    'thrs': '0.5',
    'ratio': '0.5',
    'method': '"default"',
    'uri': '"none"',
    'transpose': '0.',
    }

member_types = {
        'name': 'type',
        'char_t*': 'T_STRING',
        'uint_t': 'T_INT',
        'smpl_t': 'AUBIO_NPY_SMPL',
        }

pyfromtype_fn = {
        'smpl_t': 'PyFloat_FromDouble',
        'uint_t': 'PyLong_FromLong', # was: 'PyInt_FromLong',
        'fvec_t*': 'PyAubio_CFvecToArray',
        'fmat_t*': 'PyAubio_CFmatToArray',
        }

pytoaubio_fn = {
        'fvec_t*': 'PyAubio_ArrayToCFvec',
        'cvec_t*': 'PyAubio_PyCvecToCCvec',
        #'fmat_t*': 'PyAubio_ArrayToCFmat',
        }

newfromtype_fn = {
        'fvec_t*': 'new_py_fvec',
        'fmat_t*': 'new_py_fmat',
        'cvec_t*': 'new_py_cvec',
        }

delfromtype_fn = {
        'fvec_t*': 'Py_DECREF',
        'fmat_t*': 'Py_DECREF',
        'cvec_t*': 'Py_DECREF',
        }

param_init = {
        'char_t*': 'NULL',
        'uint_t': '0',
        'sint_t': 0,
        'smpl_t': 0.,
        'lsmp_t': 0.,
        }

pyargparse_chars = {
        'smpl_t': 'f', # if not usedouble else 'd',
        'uint_t': 'I',
        'sint_t': 'I',
        'char_t*': 's',
        'fmat_t*': 'O',
        'fvec_t*': 'O',
        'cvec_t*': 'O',
        }

objoutsize = {
        'onset': '1',
        'pitch': '1',
        'notes': '3',
        'wavetable': 'self->hop_size',
        'sampler': 'self->hop_size',
        'mfcc': 'self->n_coeffs',
        'specdesc': '1',
        'tempo': '1',
        'filterbank': 'self->n_filters',
        'tss': 'self->buf_size',
        'pitchshift': 'self->hop_size',
        'dct': 'self->size',
        }

objinputsize = {
        'mfcc': 'self->buf_size / 2 + 1',
        'notes': 'self->hop_size',
        'onset': 'self->hop_size',
        'pitch': 'self->hop_size',
        'sampler': 'self->hop_size',
        'specdesc': 'self->buf_size / 2 + 1',
        'tempo': 'self->hop_size',
        'wavetable': 'self->hop_size',
        'tss': 'self->buf_size / 2 + 1',
        'pitchshift': 'self->hop_size',
        }

def get_name(proto):
    name = proto.replace(' *', '* ').split()[1].split('(')[0]
    name = name.replace('*','')
    if name == '': raise ValueError(proto + "gave empty name")
    return name

def get_return_type(proto):
    import re
    paramregex = re.compile(r'(\w+ ?\*?).*')
    outputs = paramregex.findall(proto)
    assert len(outputs) == 1
    return outputs[0].replace(' ', '')

def split_type(arg):
    """ arg = 'foo *name' 
        return ['foo*', 'name'] """
    l = arg.split()
    type_arg = {} #'type': l[0], 'name': l[1]}
    type_arg['type'] = " ".join(l[:-1])
    type_arg['name'] = l[-1]
    # fix up type / name
    if type_arg['name'].startswith('*'):
        # ['foo', '*name'] -> ['foo*', 'name']
        type_arg['type'] += '*'
        type_arg['name'] = type_arg['name'][1:]
    if type_arg['type'].endswith(' *'):
        # ['foo *', 'name'] -> ['foo*', 'name']
        type_arg['type'] = type_arg['type'].replace(' *','*')
    if type_arg['type'].startswith('const '):
        # ['foo *', 'name'] -> ['foo*', 'name']
        type_arg['type'] = type_arg['type'].replace('const ','')
    return type_arg

def get_params(proto):
    """ get the list of parameters from a function prototype
    example: proto = "int main (int argc, char ** argv)"
    returns: ['int argc', 'char ** argv']
    """
    import re
    paramregex = re.compile(r'.*\((.*)\);')
    a = paramregex.findall(proto)[0].split(', ')
    #a = [i.replace('const ', '') for i in a]
    return a

def get_input_params(proto):
    a = get_params(proto)
    return [i.replace('const ', '') for i in a if (i.startswith('const ') or i.startswith('uint_t ') or i.startswith('smpl_t '))]

def get_output_params(proto):
    a = get_params(proto)
    return [i for i in a if not i.startswith('const ')][1:]

def get_params_types_names(proto):
    """ get the list of parameters from a function prototype
    example: proto = "int main (int argc, char ** argv)"
    returns: [['int', 'argc'], ['char **','argv']]
    """
    a = list(map(split_type, get_params(proto)))
    #print proto, a
    #import sys; sys.exit(1)
    return a

class MappedObject(object):

    def __init__(self, prototypes, usedouble = False):
        if usedouble:
            pyargparse_chars['smpl_t'] = 'd'
        self.prototypes = prototypes

        self.shortname = prototypes['shortname']
        self.longname = prototypes['longname']
        self.new_proto = prototypes['new'][0]
        self.del_proto = prototypes['del'][0]
        self.do_proto = prototypes['do'][0]
        self.input_params = get_params_types_names(self.new_proto)
        self.input_params_list = "; ".join(get_input_params(self.new_proto))
        self.outputs = get_params_types_names(self.do_proto)[2:]
        self.do_inputs = [get_params_types_names(self.do_proto)[1]]
        self.do_outputs = get_params_types_names(self.do_proto)[2:]
        struct_output_str = ["PyObject *{0[name]}; {1} c_{0[name]}".format(i, i['type'][:-1]) for i in self.do_outputs]
        if len(self.prototypes['rdo']):
            rdo_outputs = get_params_types_names(prototypes['rdo'][0])[2:]
            struct_output_str += ["PyObject *{0[name]}; {1} c_{0[name]}".format(i, i['type'][:-1]) for i in rdo_outputs]
            self.outputs += rdo_outputs
        self.struct_outputs = ";\n    ".join(struct_output_str)

        #print ("input_params: ", map(split_type, get_input_params(self.do_proto)))
        #print ("output_params", map(split_type, get_output_params(self.do_proto)))

    def gen_code(self):
        out = ""
        try:
            out += self.gen_struct()
            out += self.gen_doc()
            out += self.gen_new()
            out += self.gen_init()
            out += self.gen_del()
            out += self.gen_do()
            if len(self.prototypes['rdo']):
                self.do_proto = self.prototypes['rdo'][0]
                self.do_inputs = [get_params_types_names(self.do_proto)[1]]
                self.do_outputs = get_params_types_names(self.do_proto)[2:]
                out += self.gen_do(method='rdo')
            out += self.gen_memberdef()
            out += self.gen_set()
            out += self.gen_get()
            out += self.gen_methodef()
            out += self.gen_typeobject()
        except Exception as e:
            print ("Failed generating code for", self.shortname)
            raise
        return out

    def gen_struct(self):
        out = """
// {shortname} structure
typedef struct{{
    PyObject_HEAD
    // pointer to aubio object
    {longname} *o;
    // input parameters
    {input_params_list};
    // do input vectors
    {do_inputs_list};
    // output results
    {struct_outputs};
}} Py_{shortname};
"""
        # fmat_t* / fvec_t* / cvec_t* inputs -> full fvec_t /.. struct in Py_{shortname}
        do_inputs_list = "; ".join(get_input_params(self.do_proto)).replace('fvec_t *','fvec_t').replace('fmat_t *', 'fmat_t').replace('cvec_t *', 'cvec_t')
        return out.format(do_inputs_list = do_inputs_list, **self.__dict__)

    def gen_doc(self):
        sig = []
        for p in self.input_params:
            name = p['name']
            defval = aubiodefvalue[name].replace('"','\\\"')
            sig.append("{name}={defval}".format(defval=defval, name=name))
        out = """
#ifndef PYAUBIO_{shortname}_doc
#define PYAUBIO_{shortname}_doc "{shortname}({sig})"
#endif /* PYAUBIO_{shortname}_doc */

static char Py_{shortname}_doc[] = ""
PYAUBIO_{shortname}_doc
"";
"""
        return out.format(sig=', '.join(sig), **self.__dict__)

    def gen_new(self):
        out = """
// new {shortname}
static PyObject *
Py_{shortname}_new (PyTypeObject * pytype, PyObject * args, PyObject * kwds)
{{
    Py_{shortname} *self;
""".format(**self.__dict__)
        params = self.input_params
        for p in params:
            out += """
    {type} {name} = {defval};""".format(defval = param_init[p['type']], **p)
        plist = ", ".join(["\"%s\"" % p['name'] for p in params])
        out += """
    static char *kwlist[] = {{ {plist}, NULL }};""".format(plist = plist)
        argchars = "".join([pyargparse_chars[p['type']] for p in params])
        arglist = ", ".join(["&%s" % p['name'] for p in params])
        out += """
    if (!PyArg_ParseTupleAndKeywords (args, kwds, "|{argchars}", kwlist,
              {arglist})) {{
        return NULL;
    }}
""".format(argchars = argchars, arglist = arglist)
        out += """
    self = (Py_{shortname} *) pytype->tp_alloc (pytype, 0);
    if (self == NULL) {{
        return NULL;
    }}
""".format(**self.__dict__)
        params = self.input_params
        for p in params:
            out += self.check_valid(p)
        out += """
    return (PyObject *)self;
}
"""
        return out

    def check_valid(self, p):
        if p['type'] == 'uint_t':
            return self.check_valid_uint(p)
        if p['type'] == 'char_t*':
            return self.check_valid_char(p)
        if p['type'] == 'smpl_t':
            return self.check_valid_smpl(p)
        else:
            print ("ERROR, no idea how to check %s for validity" % p['type'])

    def check_valid_uint(self, p):
        name = p['name']
        return """
    self->{name} = {defval};
    if ((sint_t){name} > 0) {{
        self->{name} = {name};
    }} else if ((sint_t){name} < 0) {{
        PyErr_SetString (PyExc_ValueError, "can not use negative value for {name}");
        return NULL;
    }}
""".format(defval = aubiodefvalue[name], name = name)

    def check_valid_char(self, p):
        name = p['name']
        return """
    self->{name} = {defval};
    if ({name} != NULL) {{
        self->{name} = {name};
    }}
""".format(defval = aubiodefvalue[name], name = name)

    def check_valid_smpl(self, p):
        name = p['name']
        return """
    self->{name} = {defval};
    if ({name} != 0.) {{
        self->{name} = {name};
    }}
""".format(defval = aubiodefvalue[name], name = name)

    def gen_init(self):
        out = """
// init {shortname}
static int
Py_{shortname}_init (Py_{shortname} * self, PyObject * args, PyObject * kwds)
{{
""".format(**self.__dict__)
        new_name = get_name(self.new_proto)
        new_params = ", ".join(["self->%s" % s['name'] for s in self.input_params])
        out += """
  self->o = {new_name}({new_params});
""".format(new_name = new_name, new_params = new_params)
        paramchars = "%s"
        paramvals = "self->method"
        out += """
  // return -1 and set error string on failure
  if (self->o == NULL) {{
    PyErr_Format (PyExc_RuntimeError, "failed creating {shortname}");
    return -1;
  }}
""".format(paramchars = paramchars, paramvals = paramvals, **self.__dict__)
        output_create = ""
        for o in self.outputs:
            output_create += """
  self->{name} = {create_fn}({output_size});""".format(name = o['name'], create_fn = newfromtype_fn[o['type']], output_size = objoutsize[self.shortname])
        out += """
  // TODO get internal params after actual object creation?
"""
        out += """
  // create outputs{output_create}
""".format(output_create = output_create)
        out += """
  return 0;
}
"""
        return out

    def gen_memberdef(self):
        out = """
static PyMemberDef Py_{shortname}_members[] = {{
""".format(**self.__dict__)
        for p in get_params_types_names(self.new_proto):
            tmp = "  {{\"{name}\", {ttype}, offsetof (Py_{shortname}, {name}), READONLY, \"TODO documentation\"}},\n"
            pytype = member_types[p['type']]
            out += tmp.format(name = p['name'], ttype = pytype, shortname = self.shortname)
        out += """  {NULL}, // sentinel
};
"""
        return out

    def gen_del(self):
        out = """
// del {shortname}
static void
Py_{shortname}_del  (Py_{shortname} * self, PyObject * unused)
{{""".format(**self.__dict__)
        for input_param in self.do_inputs:
            if input_param['type'] == 'fmat_t *':
                out += """
  free(self->{0[name]}.data);""".format(input_param)
        for o in self.outputs:
            name = o['name']
            del_out = delfromtype_fn[o['type']]
            out += """
  if (self->{name}) {{
    {del_out}(self->{name});
  }}""".format(del_out = del_out, name = name)
        del_fn = get_name(self.del_proto)
        out += """
  if (self->o) {{
    {del_fn}(self->o);
  }}
  Py_TYPE(self)->tp_free((PyObject *) self);
}}
""".format(del_fn = del_fn)
        return out

    def gen_do(self, method = 'do'):
        out = """
// do {shortname}
static PyObject*
Pyaubio_{shortname}_{method}  (Py_{shortname} * self, PyObject * args)
{{""".format(method = method, **self.__dict__)
        input_params = self.do_inputs
        output_params = self.do_outputs
        #print input_params
        #print output_params
        out += """
    PyObject *outputs;"""
        for input_param in input_params:
            out += """
    PyObject *py_{0};""".format(input_param['name'])
        refs = ", ".join(["&py_%s" % p['name'] for p in input_params])
        pyparamtypes = "".join([pyargparse_chars[p['type']] for p in input_params])
        out += """
    if (!PyArg_ParseTuple (args, "{pyparamtypes}", {refs})) {{
        return NULL;
    }}""".format(refs = refs, pyparamtypes = pyparamtypes, **self.__dict__)
        for input_param in input_params:
            out += """

    if (!{pytoaubio}(py_{0[name]}, &(self->{0[name]}))) {{
        return NULL;
    }}""".format(input_param, pytoaubio = pytoaubio_fn[input_param['type']])
        if self.shortname in objinputsize:
            out += """

    if (self->{0[name]}.length != {expected_size}) {{
        PyErr_Format (PyExc_ValueError,
            "input size of {shortname} should be %d, not %d",
            {expected_size}, self->{0[name]}.length);
        return NULL;
    }}""".format(input_param, expected_size = objinputsize[self.shortname], **self.__dict__)
        else:
            out += """

    // TODO: check input sizes"""
        for output_param in output_params:
            out += """

    Py_INCREF(self->{0[name]});
    if (!{pytoaubio}(self->{0[name]}, &(self->c_{0[name]}))) {{
        return NULL;
    }}""".format(output_param, pytoaubio = pytoaubio_fn[output_param['type']])
        do_fn = get_name(self.do_proto)
        inputs = ", ".join(['&(self->'+p['name']+')' for p in input_params])
        c_outputs = ", ".join(["&(self->c_%s)" % p['name'] for p in self.do_outputs])
        outputs = ", ".join(["self->%s" % p['name'] for p in self.do_outputs])
        out += """

    {do_fn}(self->o, {inputs}, {c_outputs});
""".format(
        do_fn = do_fn,
        inputs = inputs, c_outputs = c_outputs,
        )
        if len(self.do_outputs) > 1:
            out += """
    outputs = PyTuple_New({:d});""".format(len(self.do_outputs))
            for i, p in enumerate(self.do_outputs):
                out += """
    PyTuple_SetItem( outputs, {i}, self->{p[name]});""".format(i = i, p = p)
        else:
            out += """
    outputs = self->{p[name]};""".format(p = self.do_outputs[0])
        out += """

    return outputs;
}}
""".format(
        outputs = outputs,
        )
        return out

    def gen_set(self):
        out = """
// {shortname} setters
""".format(**self.__dict__)
        for set_param in self.prototypes['set']:
            params = get_params_types_names(set_param)[1:]
            param = self.shortname.split('_set_')[-1]
            paramdecls = "".join(["""
   {0} {1};""".format(p['type'], p['name']) for p in params])
            method_name = get_name(set_param)
            param = method_name.split('aubio_'+self.shortname+'_set_')[-1]
            refs = ", ".join(["&%s" % p['name'] for p in params])
            paramlist = ", ".join(["%s" % p['name'] for p in params])
            if len(params):
                paramlist = "," + paramlist
            pyparamtypes = ''.join([pyargparse_chars[p['type']] for p in params])
            out += """
static PyObject *
Pyaubio_{shortname}_set_{param} (Py_{shortname} *self, PyObject *args)
{{
  uint_t err = 0;
  {paramdecls}
""".format(param = param, paramdecls = paramdecls, **self.__dict__)

            if len(refs) and len(pyparamtypes):
                out += """

  if (!PyArg_ParseTuple (args, "{pyparamtypes}", {refs})) {{
    return NULL;
  }}
""".format(pyparamtypes = pyparamtypes, refs = refs)

            out += """
  err = aubio_{shortname}_set_{param} (self->o {paramlist});

  if (err > 0) {{
    if (PyErr_Occurred() == NULL) {{
      PyErr_SetString (PyExc_ValueError, "error running aubio_{shortname}_set_{param}");
    }} else {{
      // change the RuntimeError into ValueError
      PyObject *type, *value, *traceback;
      PyErr_Fetch(&type, &value, &traceback);
      Py_XDECREF(type);
      type = PyExc_ValueError;
      Py_XINCREF(type);
      PyErr_Restore(type, value, traceback);
    }}
    return NULL;
  }}
  Py_RETURN_NONE;
}}
""".format(param = param, refs = refs, paramdecls = paramdecls,
        pyparamtypes = pyparamtypes, paramlist = paramlist, **self.__dict__)
        return out

    def gen_get(self):
        out = """
// {shortname} getters
""".format(**self.__dict__)
        for method in self.prototypes['get']:
            params = get_params_types_names(method)
            method_name = get_name(method)
            assert len(params) == 1, \
                "get method has more than one parameter %s" % params
            param = method_name.split('aubio_'+self.shortname+'_get_')[-1]
            paramtype = get_return_type(method)
            ptypeconv = pyfromtype_fn[paramtype]
            out += """
static PyObject *
Pyaubio_{shortname}_get_{param} (Py_{shortname} *self, PyObject *unused)
{{
  {ptype} {param} = aubio_{shortname}_get_{param} (self->o);
  return (PyObject *){ptypeconv} ({param});
}}
""".format(param = param, ptype = paramtype, ptypeconv = ptypeconv,
        **self.__dict__)
        return out

    def gen_methodef(self):
        out = """
static PyMethodDef Py_{shortname}_methods[] = {{""".format(**self.__dict__)
        for m in self.prototypes['set']:
            name = get_name(m)
            shortname = name.replace('aubio_%s_' % self.shortname, '')
            out += """
  {{"{shortname}", (PyCFunction) Py{name},
    METH_VARARGS, ""}},""".format(name = name, shortname = shortname)
        for m in self.prototypes['get']:
            name = get_name(m)
            shortname = name.replace('aubio_%s_' % self.shortname, '')
            out += """
  {{"{shortname}", (PyCFunction) Py{name},
    METH_NOARGS, ""}},""".format(name = name, shortname = shortname)
        for m in self.prototypes['rdo']:
            name = get_name(m)
            shortname = name.replace('aubio_%s_' % self.shortname, '')
            out += """
  {{"{shortname}", (PyCFunction) Py{name},
    METH_VARARGS, ""}},""".format(name = name, shortname = shortname)
        out += """
  {NULL} /* sentinel */
};
"""
        return out

    def gen_typeobject(self):
        return """
PyTypeObject Py_{shortname}Type = {{
  //PyObject_HEAD_INIT (NULL)
  //0,
  PyVarObject_HEAD_INIT (NULL, 0)
  "aubio.{shortname}",
  sizeof (Py_{shortname}),
  0,
  (destructor) Py_{shortname}_del,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  (ternaryfunc)Pyaubio_{shortname}_do,
  0,
  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  Py_{shortname}_doc,
  0,
  0,
  0,
  0,
  0,
  0,
  Py_{shortname}_methods,
  Py_{shortname}_members,
  0,
  0,
  0,
  0,
  0,
  0,
  (initproc) Py_{shortname}_init,
  0,
  Py_{shortname}_new,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
}};
""".format(**self.__dict__)



"""
This file imports test methods from different testing modules, in this
order:

    - try importing 'pytest'
    - if it fails, fallback to 'numpy.testing'

Nose2 support was removed because of lacking assertWarns on py2.7.

"""

import sys

_has_pytest = False

# check if we have pytest
try:
    import pytest
    parametrize = pytest.mark.parametrize
    assert_raises = pytest.raises
    assert_warns = pytest.warns
    skipTest = pytest.skip
    _has_pytest = True
    def run_module_suite():
        import sys, pytest
        pytest.main(sys.argv)
except:
    pass

# otherwise fallback on numpy.testing
if not _has_pytest:
    from numpy.testing import dec, assert_raises, assert_warns
    from numpy.testing import SkipTest
    parametrize = dec.parametrize
    def skipTest(msg):
        raise SkipTest(msg)
    from numpy.testing import run_module_suite

# always use numpy's assert_equal
import numpy
assert_equal = numpy.testing.assert_equal



#! /usr/bin/env python

""" this file was written by Paul Brossier
  it is released under the GNU/GPL license.
"""

import sys
from aubio.cmd import AubioArgumentParser, _cut_slice

def aubio_cut_parser():
    parser = AubioArgumentParser()
    parser.add_input()
    parser.add_argument("-O", "--onset-method",
            action="store", dest="onset_method", default='default',
            metavar = "<onset_method>",
            help="onset detection method [default=default] \
                    complexdomain|hfc|phase|specdiff|energy|kl|mkl")
    # cutting methods
    parser.add_argument("-b", "--beat",
            action="store_true", dest="beat", default=False,
            help="slice at beat locations")
    """
    parser.add_argument("-S", "--silencecut",
            action="store_true", dest="silencecut", default=False,
            help="use silence locations")
    parser.add_argument("-s", "--silence",
            metavar = "<value>",
            action="store", dest="silence", default=-70,
            help="silence threshold [default=-70]")
            """
    # algorithm parameters
    parser.add_buf_hop_size()
    parser.add_argument("-t", "--threshold", "--onset-threshold",
            metavar = "<threshold>", type=float,
            action="store", dest="threshold", default=0.3,
            help="onset peak picking threshold [default=0.3]")
    parser.add_argument("-c", "--cut",
            action="store_true", dest="cut", default=False,
            help="cut input sound file at detected labels")
    parser.add_minioi()

    """
    parser.add_argument("-D", "--delay",
            action = "store", dest = "delay", type = float,
            metavar = "<seconds>", default=0,
            help="number of seconds to take back [default=system]\
                    default system delay is 3*hopsize/samplerate")
    parser.add_argument("-C", "--dcthreshold",
            metavar = "<value>",
            action="store", dest="dcthreshold", default=1.,
            help="onset peak picking DC component [default=1.]")
    parser.add_argument("-L", "--localmin",
            action="store_true", dest="localmin", default=False,
            help="use local minima after peak detection")
    parser.add_argument("-d", "--derivate",
            action="store_true", dest="derivate", default=False,
            help="derivate onset detection function")
    parser.add_argument("-z", "--zerocross",
            metavar = "<value>",
            action="store", dest="zerothres", default=0.008,
            help="zero-crossing threshold for slicing [default=0.00008]")
    # plotting functions
    parser.add_argument("-p", "--plot",
            action="store_true", dest="plot", default=False,
            help="draw plot")
    parser.add_argument("-x", "--xsize",
            metavar = "<size>",
            action="store", dest="xsize", default=1.,
            type=float, help="define xsize for plot")
    parser.add_argument("-y", "--ysize",
            metavar = "<size>",
            action="store", dest="ysize", default=1.,
            type=float, help="define ysize for plot")
    parser.add_argument("-f", "--function",
            action="store_true", dest="func", default=False,
            help="print detection function")
    parser.add_argument("-n", "--no-onsets",
            action="store_true", dest="nplot", default=False,
            help="do not plot detected onsets")
    parser.add_argument("-O", "--outplot",
            metavar = "<output_image>",
            action="store", dest="outplot", default=None,
            help="save plot to output.{ps,png}")
    parser.add_argument("-F", "--spectrogram",
            action="store_true", dest="spectro", default=False,
            help="add spectrogram to the plot")
    """
    parser.add_slicer_options()
    parser.add_verbose_help()
    return parser


def _cut_analyze(options):
    hopsize = options.hop_size
    bufsize = options.buf_size
    samplerate = options.samplerate
    source_uri = options.source_uri

    # analyze pass
    from aubio import onset, tempo, source

    s = source(source_uri, samplerate, hopsize)
    if samplerate == 0:
        samplerate = s.samplerate
        options.samplerate = samplerate

    if options.beat:
        o = tempo(options.onset_method, bufsize, hopsize,
                samplerate=samplerate)
    else:
        o = onset(options.onset_method, bufsize, hopsize,
                samplerate=samplerate)
        if options.minioi:
            if options.minioi.endswith('ms'):
                o.set_minioi_ms(int(options.minioi[:-2]))
            elif options.minioi.endswith('s'):
                o.set_minioi_s(int(options.minioi[:-1]))
            else:
                o.set_minioi(int(options.minioi))
    o.set_threshold(options.threshold)

    timestamps = []
    total_frames = 0
    while True:
        samples, read = s()
        if o(samples):
            timestamps.append(o.get_last())
            if options.verbose:
                print("%.4f" % o.get_last_s())
        total_frames += read
        if read < hopsize:
            break
    del s
    return timestamps, total_frames

def main():
    parser = aubio_cut_parser()
    options = parser.parse_args()
    if not options.source_uri and not options.source_uri2:
        sys.stderr.write("Error: no file name given\n")
        parser.print_help()
        sys.exit(1)
    elif options.source_uri2 is not None:
        options.source_uri = options.source_uri2

    # analysis
    timestamps, total_frames = _cut_analyze(options)

    # print some info
    duration = float(total_frames) / float(options.samplerate)
    base_info = '%(source_uri)s' % {'source_uri': options.source_uri}
    base_info += ' (total %(duration).2fs at %(samplerate)dHz)\n' % \
            {'duration': duration, 'samplerate': options.samplerate}

    info = "found %d timestamps in " % len(timestamps)
    info += base_info
    sys.stderr.write(info)

    if options.cut:
        _cut_slice(options, timestamps)
        info = "created %d slices from " % len(timestamps)
        info += base_info
        sys.stderr.write(info)



#! /usr/bin/env python

from aubio import miditofreq
from numpy import arange

upsampling = 100.
midi = arange(-10, 148 * upsampling)
midi /= upsampling
freq = miditofreq(midi)

from matplotlib import pyplot as plt

ax = plt.axes()
ax.semilogy(midi, freq, '.')
ax.set_xlabel('midi note')
ax.set_ylabel('frequency (Hz)')
plt.show()



#! /usr/bin/env python

import sys
from math import pi, e
from aubio import sink, float_type
from numpy import arange, sin, exp, zeros

if len(sys.argv) < 2:
    print('usage: %s <outputfile> [samplerate]' % sys.argv[0])
    sys.exit(1)

samplerate = 44100 # samplerate in Hz
if len( sys.argv ) > 2: samplerate = int(sys.argv[2])

pitch = 2200            # in Hz
blocksize = 256         # in samples
duration = 0.02         # in seconds

twopi = pi * 2.

duration = int ( samplerate * duration ) # convert to samples
attack = int (samplerate * .001 )
decay = .5

period = float(samplerate) /  pitch
# create a sine lookup table
tablelen = 1000
sinetable = arange(tablelen + 1, dtype = float_type)
sinetable = 0.7 * sin(twopi * sinetable/tablelen)
sinetone = zeros((duration,), dtype = float_type)

# compute sinetone at floating point period
for i in range(duration):
    x = int((i % period) / float(period) * tablelen)
    idx = int(x)
    frac = x - idx
    a = sinetable[idx]
    b = sinetable[idx + 1]
    sinetone[i] = a + frac * (b -a)

# apply some envelope
float_ramp = arange(duration, dtype = float_type)
sinetone *= exp( - e * float_ramp / duration / decay)
sinetone[:attack] *= exp( e * ( float_ramp[:attack] / attack - 1 ) )

if 1:
    import matplotlib.pyplot as plt
    plt.plot(sinetone)
    plt.show()

my_sink = sink(sys.argv[1], samplerate)

total_frames = 0
while total_frames + blocksize < duration:
    my_sink(sinetone[total_frames:total_frames+blocksize], blocksize)
    total_frames += blocksize
my_sink(sinetone[total_frames:duration], duration - total_frames)



#! /usr/bin/env python

from numpy.testing import TestCase
from _tools import parametrize, skipTest
import numpy as np
import aubio

class aubio_pitchshift(TestCase):

    def setUp(self):
        try:
            self.o = aubio.pitchshift(hop_size = 128)
        except RuntimeError as e:
            self.skipTest("creating aubio.pitchshift {}".format(e))

    def test_default_creation(self):
        self.assertEqual(self.o.get_pitchscale(), 1)
        self.assertEqual(self.o.get_transpose(), 0)

    def test_on_zeros(self):
        test_length = self.o.hop_size * 100
        read = 0
        # test on zeros
        vec = aubio.fvec(self.o.hop_size)
        transpose_range = 24
        while read < test_length:
            # transpose the samples
            out = self.o(vec)
            self.assertTrue((out == 0).all())
            # position in the file (between 0. and 1.)
            percent_read = read / float(test_length)
            # variable transpose rate (in semitones)
            transpose = 2 * transpose_range * percent_read - transpose_range
            # set transpose rate
            self.o.set_transpose(transpose)
            read += len(vec)

    def test_on_ones(self):
        test_length = self.o.hop_size * 100
        read = 0
        # test on zeros
        vec = aubio.fvec(self.o.hop_size) + 1
        transpose_range = 1.24
        while read < test_length:
            # transpose the samples
            out = self.o(vec)
            # position in the file (between 0. and 1.)
            percent_read = read / float(test_length)
            # variable transpose rate (in semitones)
            transpose = 2 * transpose_range * percent_read - transpose_range
            # set transpose rate
            self.o.set_transpose(transpose)
            read += len(vec)

    def test_transpose_too_high(self):
        with self.assertRaises(ValueError):
            self.o.set_transpose(24.3)

    def test_transpose_too_low(self):
        with self.assertRaises(ValueError):
            self.o.set_transpose(-24.3)

class aubio_pitchshift_wrong_params(TestCase):

    def test_wrong_transpose(self):
        with self.assertRaises(RuntimeError):
            aubio.pitchshift("default", -123)

class Test_aubio_pitchshift_testruns(object):

    run_args = ['mode', 'pitchscale', 'hop_size', 'samplerate']
    run_values = [
            ("default",     1.2,  128,  44100),
            ("crispness:0", 0.43,  64,   8000),
            ("crispness:3", 0.53, 256,   8000),
            ("crispness:3", 1.53, 512,   8000),
            ("crispness:6", 2.3, 4096, 192000),
            ]

    @parametrize(run_args, run_values)
    def test_run_with_params(self, mode, pitchscale, hop_size, samplerate):
        try:
            self.o = aubio.pitchshift(mode, pitchscale, hop_size, samplerate)
        except RuntimeError as e:
            skipTest("failed creating pitchshift ({})".format(e))
        test_length = self.o.hop_size * 50
        read = 0
        # test on random
        vec = np.random.rand(self.o.hop_size).astype(aubio.float_type)
        transpose_range = self.o.get_transpose()
        while read < test_length:
            # transpose the samples
            out = self.o(vec)
            # position in the file (between 0. and 1.)
            percent_read = read / float(test_length)
            # variable transpose rate (in semitones)
            transpose =  transpose_range - 2 * transpose_range * percent_read
            # set transpose rate
            self.o.set_transpose(transpose)
            read += len(vec)

if __name__ == '__main__':
    from _tools import run_module_suite
    run_module_suite()



#! /usr/bin/env python

import sys
from aubio import source, pvoc, mfcc
from numpy import vstack, zeros, diff

n_filters = 40              # must be 40 for mfcc
n_coeffs = 13

if len(sys.argv) < 2:
    print("Usage: %s <source_filename> [samplerate] [win_s] [hop_s] [mode]" % sys.argv[0])
    print("  where [mode] can be 'delta' or 'ddelta' for first and second derivatives")
    sys.exit(1)

source_filename = sys.argv[1]

if len(sys.argv) > 2: samplerate = int(sys.argv[2])
else: samplerate = 0
if len(sys.argv) > 3: win_s = int(sys.argv[3])
else: win_s = 512
if len(sys.argv) > 4: hop_s = int(sys.argv[4])
else: hop_s = win_s // 4
if len(sys.argv) > 5: mode = sys.argv[5]
else: mode = "default"

samplerate = 0
if len( sys.argv ) > 2: samplerate = int(sys.argv[2])

s = source(source_filename, samplerate, hop_s)
samplerate = s.samplerate
p = pvoc(win_s, hop_s)
m = mfcc(win_s, n_filters, n_coeffs, samplerate)

mfccs = zeros([n_coeffs,])
frames_read = 0
while True:
    samples, read = s()
    spec = p(samples)
    mfcc_out = m(spec)
    mfccs = vstack((mfccs, mfcc_out))
    frames_read += read
    if read < hop_s: break

# do plotting
from numpy import arange
from demo_waveform_plot import get_waveform_plot
from demo_waveform_plot import set_xlabels_sample2time
import matplotlib.pyplot as plt

fig = plt.figure()
plt.rc('lines',linewidth='.8')
wave = plt.axes([0.1, 0.75, 0.8, 0.19])

get_waveform_plot( source_filename, samplerate, block_size = hop_s, ax = wave)
wave.xaxis.set_visible(False)
wave.yaxis.set_visible(False)

# compute first and second derivatives
if mode in ["delta", "ddelta"]:
    mfccs = diff(mfccs, axis = 0)
if mode == "ddelta":
    mfccs = diff(mfccs, axis = 0)

all_times = arange(mfccs.shape[0]) * hop_s
n_coeffs = mfccs.shape[1]
for i in range(n_coeffs):
    ax = plt.axes ( [0.1, 0.75 - ((i+1) * 0.65 / n_coeffs),  0.8, 0.65 / n_coeffs], sharex = wave )
    ax.xaxis.set_visible(False)
    ax.set_yticks([])
    ax.set_ylabel('%d' % i)
    ax.plot(all_times, mfccs.T[i])

# add time to the last axis
set_xlabels_sample2time( ax, frames_read, samplerate)

#plt.ylabel('spectral descriptor value')
ax.xaxis.set_visible(True)
title = 'MFCC for %s' % source_filename
if mode == "delta": title = mode + " " + title
elif mode == "ddelta": title = "double-delta" + " " + title
wave.set_title(title)
plt.show()



#! /usr/bin/env python

import os
import re
import glob
import struct
import numpy as np
from tempfile import mkstemp

DEFAULT_SOUND = '22050Hz_5s_brownnoise.wav'

def is32bit():
    return struct.calcsize("P") * 8 == 32

def array_from_text_file(filename, dtype = 'float'):
    realpathname = os.path.join(os.path.dirname(__file__), filename)
    return np.loadtxt(realpathname, dtype = dtype)

def list_all_sounds(rel_dir):
    datadir = os.path.join(os.path.dirname(__file__), rel_dir)
    return glob.glob(os.path.join(datadir,'*.*'))

def get_default_test_sound(TestCase, rel_dir = 'sounds'):
    all_sounds = list_all_sounds(rel_dir)
    if len(all_sounds) == 0:
        TestCase.skipTest("please add some sounds in \'python/tests/sounds\'")
    else:
        default_sound = all_sounds[0]
        if DEFAULT_SOUND in map(os.path.basename, all_sounds):
            while os.path.basename(default_sound) != DEFAULT_SOUND:
                default_sound = all_sounds.pop(0)
        return default_sound

def get_tmp_sink_path():
    fd, path = mkstemp()
    os.close(fd)
    return path

def del_tmp_sink_path(path):
    try:
        os.unlink(path)
    except WindowsError as e:
        # removing the temporary directory sometimes fails on windows
        import warnings
        errmsg = "failed deleting temporary file {:s} ({:s})"
        warnings.warn(UserWarning(errmsg.format(path, repr(e))))

def array_from_yaml_file(filename):
    import yaml
    f = open(filename)
    yaml_data = yaml.safe_load(f)
    f.close()
    return yaml_data

def count_samples_in_file(file_path):
    from aubio import source
    hopsize = 256
    s = source(file_path, 0, hopsize)
    total_frames = 0
    while True:
        _, read = s()
        total_frames += read
        if read < hopsize: break
    return total_frames

def count_samples_in_directory(samples_dir):
    total_frames = 0
    for f in os.walk(samples_dir):
        if len(f[2]):
            for each in f[2]:
                file_path = os.path.join(f[0], each)
                if file_path:
                    total_frames += count_samples_in_file(file_path)
    return total_frames

def count_files_in_directory(samples_dir):
    total_files = 0
    for f in os.walk(samples_dir):
        if len(f[2]):
            for each in f[2]:
                file_path = os.path.join(f[0], each)
                if file_path:
                    total_files += 1
    return total_files

def parse_file_samplerate(soundfile):
    samplerate = None
    # parse samplerate - match numbers followed by Hz in filename, with optional path separator
    re_sr = re.compile(r'[/\\]?([0-9]{4,})Hz[^0-9]')
    match_samplerate = re_sr.findall(soundfile)
    if match_samplerate:
        samplerate = int(match_samplerate[0])
    else:
        import warnings
        warnings.warn(UserWarning(f"could not parse samplerate for {soundfile}"))
    return samplerate
